# awl.admintools.py
from django.contrib.admin import ModelAdmin
from django.contrib.admin.utils import lookup_field
from django.template import Context, Template
from django.urls import reverse
from django.utils.html import format_html

from awl.utils import get_obj_attr

# ============================================================================

def admin_obj_link(obj, display=''):
    """Returns a link to the django admin change list with a filter set to
    only the object given.

    :param obj:
        Object to create the admin change list display link for
    :param display:
        Text to display in the link.  Defaults to string call of the object
    :returns:
        Text containing HTML for a link
    """
    # get the url for the change list for this object
    url = reverse('admin:%s_%s_changelist' % (obj._meta.app_label,
        obj._meta.model_name))
    url += '?id__exact=%s' % obj.id

    text = str(obj)
    if display:
        text = display

    return format_html('<a href="{}">{}</a>', url, text)


def admin_obj_attr(obj, attr):
    """A safe version of :func:``utils.get_obj_attr`` that returns and empty
    string in the case of an exception or an empty object
    """
    try:
        field_obj = get_obj_attr(obj, attr)
        if not field_obj:
            return ''
    except AttributeError:
        return ''

    return field_obj


def _obj_display(obj, display=''):
    """Returns string representation of an object, either the default or based
    on the display template passed in.
    """
    result = ''
    if not display:
        result = str(obj)
    else:
        template = Template(display)
        context = Context({'obj':obj})
        result = template.render(context)

    return result


def make_admin_obj_mixin(name):
    """This method dynamically creates a mixin to be used with your 
    :class:`ModelAdmin` classes.  The mixin provides utility methods that can
    be referenced in side of the admin object's ``list_display`` and other
    similar attributes.
    
    :param name:
        Each usage of the mixin must be given a unique name for the mixin class
        being created
    :returns:
        Dynamically created mixin class

    The created class supports the following methods:

    .. code-block:: python

        add_obj_ref(funcname, attr, [title, display])


    Django admin ``list_display`` does not support the double underscore
    semantics of object references.  This method adds a function to the mixin
    that returns the ``str(obj)`` value from object relations.

    :param funcname:
        Name of the function to be added to the mixin.  In the admin class
        object that includes the mixin, this name is used in the
        ``list_display`` tuple.
    :param attr:
        Name of the attribute to dereference from the corresponding object,
        i.e. what will be dereferenced.  This name supports double underscore
        object link referencing for ``models.ForeignKey`` members.
    :param title:
        Title for the column of the django admin table.  If not given it
        defaults to a capitalized version of ``attr``
    :param display:
        What to display as the text in the column.  If not given it defaults
        to the string representation of the object for the row: ``str(obj)`` .
        This parameter supports django templating, the context for which
        contains a dictionary key named "obj" with the value being the object
        for the row.

    .. code-block:: python

        add_obj_link(funcname, attr, [title, display])


    This method adds a function to the mixin that returns a link to a django
    admin change list page for the member attribute of the object being
    displayed.

    :param funcname:
        Name of the function to be added to the mixin.  In the admin class
        object that includes the mixin, this name is used in the
        ``list_display`` tuple.
    :param attr:
        Name of the attribute to dereference from the corresponding object,
        i.e. what will be lined to.  This name supports double underscore
        object link referencing for ``models.ForeignKey`` members.
    :param title:
        Title for the column of the django admin table.  If not given it
        defaults to a capitalized version of ``attr``
    :param display:
        What to display as the text for the link being shown.  If not given it
        defaults to the string representation of the object for the row: 
        ``str(obj)`` .  This parameter supports django templating, the context
        for which contains a dictionary key named "obj" with the value being
        the object for the row.

    Example usage:

    .. code-block:: python

        # ---- models.py file ----
        class Author(models.Model):
            name = models.CharField(max_length=100)


        class Book(models.Model):
            title = models.CharField(max_length=100)
            author = models.ForeignKey(Author, on_delete=models.CASCADE)


    .. code-block:: python

        # ---- admin.py file ----
        @admin.register(Author)
        class Author(admin.ModelAdmin):
            list_display = ('name', )


        mixin = make_admin_obj_mixin('BookMixin')
        mixin.add_obj_link('show_author', 'Author', 'Our Authors',
            '{{obj.name}} (id={{obj.id}})')

        @admin.register(Book)
        class BookAdmin(admin.ModelAdmin, mixin):
            list_display = ('name', 'show_author')


    A sample django admin page for "Book" would have the table:

    +---------------------------------+------------------------+
    | Name                            | Our Authors            |
    +=================================+========================+
    | Hitchhikers Guide To The Galaxy | *Douglas Adams (id=1)* |
    +---------------------------------+------------------------+
    | War and Peace                   | *Tolstoy (id=2)*       |
    +---------------------------------+------------------------+
    | Dirk Gently                     | *Douglas Adams (id=1)* |
    +---------------------------------+------------------------+


    Each of the *items* in the "Our Authors" column would be a link to the
    django admin change list for the "Author" object with a filter set to show
    just the object that was clicked.  For example, if you clicked "Douglas
    Adams (id=1)" you would be taken to the Author change list page filtered
    just for Douglas Adams books.

    The ``add_obj_ref`` method is similar to the above, but instead of showing
    links, it just shows text and so can be used for view-only attributes of
    dereferenced objects.
    """
    @classmethod
    def add_obj_link(cls, funcname, attr, title='', display=''):
        if not title:
            title = attr.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # referenced the inner function won't see it, so assign it for use
        _display = display

        def _link(self, obj):
            field_obj = admin_obj_attr(obj, attr)
            if not field_obj:
                return ''

            text = _obj_display(field_obj, _display)
            return admin_obj_link(field_obj, text)
        _link.short_description = title
        _link.allow_tags = True
        _link.admin_order_field = attr

        setattr(cls, funcname, _link)

    @classmethod
    def add_obj_ref(cls, funcname, attr, title='', display=''):
        if not title:
            title = attr.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # referenced the inner function won't see it, so assign it for use
        _display = display

        def _ref(self, obj):
            field_obj = admin_obj_attr(obj, attr)
            if not field_obj:
                return ''

            return _obj_display(field_obj, _display)
        _ref.short_description = title
        _ref.allow_tags = True
        _ref.admin_order_field = attr

        setattr(cls, funcname, _ref)

    klass = type(name, (), {})
    klass.add_obj_link = add_obj_link
    klass.add_obj_ref = add_obj_ref
    return klass

# ============================================================================
# Newer Admin Object Mechanism

klass_count = 0

class FancyModelAdmin(ModelAdmin):
    """A replacement for :class:`admin.ModelAdmin` which provides additional
    methods for improving how the ``list_display`` attribute works. 

    This class should not be instantiated directly, instead call 
    :func:`fancy_modeladmin`.
    """
    list_display = []

    @classmethod
    def add_displays(cls, *args):
        """Each arg is added to the ``list_display`` property without any
        extra wrappers, using only the regular django functionality"""
        for arg in args:
            cls.list_display.append(arg)

    @classmethod
    def add_display(cls, attr, title=''):
        """Adds a ``list_display`` property without any extra wrappers,
        similar to :func:`add_displays`, but can also change the title.

        :param attr:
            Name of the attribute to add to the display

        :param title:
            Title for the column of the django admin table.  If not given it
            defaults to a capitalized version of ``attr``
        """
        global klass_count
        klass_count += 1
        fn_name = 'dyn_fn_%d' % klass_count
        cls.list_display.append(fn_name)

        if not title:
            title = attr.capitalize()

        def _ref(self, obj):
            # use the django mechanism for field value lookup
            _, _, value = lookup_field(attr, obj, cls)
            return value
        _ref.short_description = title
        _ref.allow_tags = True
        _ref.admin_order_field = attr

        setattr(cls, fn_name, _ref)

    @classmethod
    def add_link(cls, attr, title='', display=''):
        """Adds a ``list_display`` attribute that appears as a link to the
        django admin change page for the type of object being shown. Supports
        double underscore attribute name dereferencing.

        :param attr:
            Name of the attribute to dereference from the corresponding
            object, i.e. what will be lined to.  This name supports double
            underscore object link referencing for ``models.ForeignKey``
            members.

        :param title:
            Title for the column of the django admin table.  If not given it
            defaults to a capitalized version of ``attr``

        :param display:
            What to display as the text for the link being shown.  If not
            given it defaults to the string representation of the object for
            the row: ``str(obj)`` .  This parameter supports django
            templating, the context for which contains a dictionary key named
            "obj" with the value being the object for the row.

        Example usage:

        .. code-block:: python

            # ---- admin.py file ----

            base = fancy_modeladmin('id')
            base.add_link('author', 'Our Authors',
                '{{obj.name}} (id={{obj.id}})')

            @admin.register(Book)
            class BookAdmin(base):
                pass

        The django admin change page for the Book class would have a column
        for "id" and another titled "Our Authors". The "Our Authors" column
        would have a link for each Author object referenced by "book.author".
        The link would go to the Author django admin change listing. The
        display of the link would be the name of the author with the id in
        brakcets, e.g. "Douglas Adams (id=42)"
        """
        global klass_count
        klass_count += 1
        fn_name = 'dyn_fn_%d' % klass_count
        cls.list_display.append(fn_name)

        if not title:
            title = attr.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # referenced the inner function won't see it, so assign it for use
        _display = display

        def _link(self, obj):
            field_obj = admin_obj_attr(obj, attr)
            if not field_obj:
                return ''

            text = _obj_display(field_obj, _display)
            return admin_obj_link(field_obj, text)
        _link.short_description = title
        _link.allow_tags = True
        _link.admin_order_field = attr

        setattr(cls, fn_name, _link)

    @classmethod
    def add_object(cls, attr, title='', display=''):
        """Adds a ``list_display`` attribute showing an object.  Supports
        double underscore attribute name dereferencing.

        :param attr:
            Name of the attribute to dereference from the corresponding
            object, i.e. what will be lined to.  This name supports double
            underscore object link referencing for ``models.ForeignKey``
            members.

        :param title:
            Title for the column of the django admin table.  If not given it
            defaults to a capitalized version of ``attr``

        :param display:
            What to display as the text for the link being shown.  If not
            given it defaults to the string representation of the object for
            the row: ``str(obj)``.  This parameter supports django templating,
            the context for which contains a dictionary key named "obj" with
            the value being the object for the row.
        """
        global klass_count
        klass_count += 1
        fn_name = 'dyn_fn_%d' % klass_count
        cls.list_display.append(fn_name)

        if not title:
            title = attr.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # referenced the inner function won't see it, so assign it for use
        _display = display

        def _ref(self, obj):
            field_obj = admin_obj_attr(obj, attr)
            if not field_obj:
                return ''

            return _obj_display(field_obj, _display)
        _ref.short_description = title
        _ref.allow_tags = True
        _ref.admin_order_field = attr

        setattr(cls, fn_name, _ref)

    @classmethod
    def add_formatted_field(cls, field, format_string, title=''):
        """Adds a ``list_display`` attribute showing a field in the object
        using a python %formatted string.

        :param field:
            Name of the field in the object.

        :param format_string:
            A old-style (to remain python 2.x compatible) % string formatter
            with a single variable reference. The named ``field`` attribute
            will be passed to the formatter using the "%" operator. 

        :param title:
            Title for the column of the django admin table.  If not given it
            defaults to a capitalized version of ``field``
        """
        global klass_count
        klass_count += 1
        fn_name = 'dyn_fn_%d' % klass_count
        cls.list_display.append(fn_name)

        if not title:
            title = field.capitalize()

        # python scoping is a bit weird with default values, if it isn't
        # referenced the inner function won't see it, so assign it for use
        _format_string = format_string

        def _ref(self, obj):
            return _format_string % getattr(obj, field)
        _ref.short_description = title
        _ref.allow_tags = True
        _ref.admin_order_field = field

        setattr(cls, fn_name, _ref)



def fancy_modeladmin(*args):
    """Returns a new copy of a :class:`FancyModelAdmin` class (a class, not
    an instance!). This can then be inherited from when declaring a model
    admin class. The :class:`FancyModelAdmin` class has additional methods
    for managing the ``list_display`` attribute.

    :param ``*args``: [optional] any arguments given will be added to the
        ``list_display`` property using regular django ``list_display``
        functionality.

    This function is meant as a replacement for :func:`make_admin_obj_mixin`,
    it does everything the old one does with fewer bookkeeping needs for the
    user as well as adding functionality.

    Example usage:

    .. code-block:: python

        # ---- models.py file ----
        class Author(models.Model):
            name = models.CharField(max_length=100)


        class Book(models.Model):
            title = models.CharField(max_length=100)
            author = models.ForeignKey(Author, on_delete=models.CASCADE)


    .. code-block:: python

        # ---- admin.py file ----
        @admin.register(Author)
        class Author(admin.ModelAdmin):
            list_display = ('name', )


        base = fany_list_display_modeladmin()
        base.add_displays('id', 'name')
        base.add_obj_link('author', 'Our Authors',
            '{{obj.name}} (id={{obj.id}})')

        @admin.register(Book)
        class BookAdmin(base):
            list_display = ('name', 'show_author')


    A sample django admin page for "Book" would have the table:

    +----+---------------------------------+------------------------+
    | ID | Name                            | Our Authors            |
    +====+=================================+========================+
    |  1 | Hitchhikers Guide To The Galaxy | *Douglas Adams (id=1)* |
    +----+---------------------------------+------------------------+
    |  2 | War and Peace                   | *Tolstoy (id=2)*       |
    +----+---------------------------------+------------------------+
    |  3 | Dirk Gently                     | *Douglas Adams (id=1)* |
    +----+---------------------------------+------------------------+


    See :class:`FancyModelAdmin` for a full list of functionality
    provided by the returned base class.
    """
    global klass_count

    klass_count += 1
    name = 'DynamicAdminClass%d' % klass_count

    # clone the admin class
    klass = type(name, (FancyModelAdmin,), {})
    klass.list_display = []
    if len(args) > 0:
        klass.add_displays(*args)

    return klass



# awl.managment.commands.create_test_admin.py
#
# Creates an admin user, done here so it can be automated as part of the reset
# script

from django.core.management.base import BaseCommand
from awl.waelsteng import create_admin

class Command(BaseCommand):
    """Creates a test admin user account by calling
    :func:'awl.waelsteng.create_admin'"""

    help = ('Creates a test admin user account by calling '
        'awl.waelsteng.create_admin')

    def handle(self, *args, **options):
        create_admin()



# awl.tests.test_commands.py
import os, mock

from django.conf import settings
from django.contrib.auth.models import User
from django.core.management import call_command, CommandError
from django.test import TestCase, override_settings

from context_temp import temp_directory
from waelstow import capture_stdout

# ============================================================================

class CommandTests(TestCase):
    def test_create_admin(self):
        call_command('create_test_admin')

        # verify user created by script exists
        User.objects.get(username='admin')

    def test_run_script(self):
        filename = os.path.abspath(os.path.join(
            os.path.dirname(__file__), 'data/runme.py'))

        call_command('run_script', filename)

        # verify user created by script exists
        User.objects.get(username='fakeuser')

    @mock.patch('os.remove')
    def test_wipe_migrations(self, mocked_remove):
        call_command('wipe_migrations')

        self.assertTrue(mocked_remove.called)

    def test_print_setting(self):
        with override_settings(FOO='thingy'):
            with capture_stdout() as capture:
                call_command('print_setting', 'FOO')

            self.assertEqual(capture.getvalue(), 'thingy\n')

        with override_settings(BAR=['one', 'two']):
            with capture_stdout() as capture:
                call_command('print_setting', 'BAR')

            self.assertEqual(capture.getvalue(), 'one two\n')

        with override_settings(BAZ=('three', 'four')):
            with capture_stdout() as capture:
                call_command('print_setting', 'BAZ')

            self.assertEqual(capture.getvalue(), 'three four\n')

    def test_create_cmd(self):
        with temp_directory(path=settings.BASE_DIR) as td:
            app_name = os.path.basename(td)
            mgm_dir = os.path.join(td, 'management')
            mgm_init = os.path.join(mgm_dir, '__init__.py')
            cmd_dir = os.path.join(mgm_dir, 'commands')
            cmd_init = os.path.join(cmd_dir, '__init__.py')
            cmd_file = os.path.join(cmd_dir, 'cmd.py')
            named_file = os.path.join(cmd_dir, 'named.py')

            # test from scratch
            call_command('create_cmd', app_name)
            self.assertTrue(os.path.isdir(mgm_dir))
            self.assertTrue(os.path.isfile(mgm_init))
            self.assertTrue(os.path.isdir(cmd_dir))
            self.assertTrue(os.path.isfile(cmd_init))
            self.assertTrue(os.path.isfile(cmd_file))

            # test with named file on existing structure
            call_command('create_cmd', app_name, 'named')
            self.assertTrue(os.path.isfile(named_file))

            # test bad app name
            with self.assertRaises(CommandError) as cm:
                call_command('create_cmd', 'foo')

            self.assertIn('No such app', cm.exception.args[0])

            # test no file over-write
            with self.assertRaises(CommandError) as cm:
                call_command('create_cmd', app_name)

            self.assertIn('already existed', cm.exception.args[0])



from django.db import models
from django.utils.encoding import python_2_unicode_compatible

from awl.absmodels import ValidatingMixin
from awl.rankedmodel.models import RankedModel

# ============================================================================
# Waelsteng Models
# ============================================================================

class Link(models.Model):
    url = models.CharField(max_length=80)
    text = models.CharField(max_length=80)


class Validator(ValidatingMixin, models.Model):
    counter = models.IntegerField(default=0)

    def full_clean(self):
        super(Validator, self).full_clean()
        self.counter += 1

# ============================================================================
# Admintools Models 
# ============================================================================

@python_2_unicode_compatible
class Author(models.Model):
    name = models.CharField(max_length=20)

    def __str__(self):
        return 'Author(id=%s %s)' % (self.id, self.name)


@python_2_unicode_compatible
class Book(models.Model):
    name = models.CharField(max_length=20)
    author = models.ForeignKey(Author, null=True, blank=True,
        on_delete=models.CASCADE)

    def __str__(self):
        return 'Book(id=%s %s)' % (self.id, self.name)

    @property
    def classname(self):
        return 'Book'


class Chapter(models.Model):
    name = models.CharField(max_length=20)
    book = models.ForeignKey(Book, null=True, blank=True,
        on_delete=models.CASCADE)

# ----------------------------------------------------------------------------

@python_2_unicode_compatible
class VehicleMake(models.Model):
    name = models.CharField(max_length=20)

    def __str__(self):
        return 'VehicleMake(id=%s %s)' % (self.id, self.name)


@python_2_unicode_compatible
class VehicleModel(models.Model):
    name = models.CharField(max_length=20)
    year = models.PositiveIntegerField()
    vehiclemake = models.ForeignKey(VehicleMake, null=True, blank=True,
        on_delete=models.CASCADE)

    def __str__(self):
        return 'VehicleModel(id=%s %s)' % (self.id, self.name)

    @property
    def fullname(self):
        return '%s %s' % (self.vehiclemake.name, self.name)


class Driver(models.Model):
    name = models.CharField(max_length=20)
    vehiclemodel = models.ForeignKey(VehicleModel, null=True, blank=True,
        on_delete=models.CASCADE)
    rating = models.FloatField(null=True)


# ============================================================================
# RankedModel Models 

class Alone(RankedModel):
    name = models.CharField(max_length=1)

    def __init__(self, *args, **kwargs):
        # ignore fake group entry to make the constructors the same for both
        # test models
        kwargs.pop('group', None)
        super(Alone, self).__init__(*args, **kwargs)


class Grouped(RankedModel):
    group = models.CharField(max_length=1)
    name = models.CharField(max_length=1)

    def grouped_filter(self):
        return Grouped.objects.filter(group=self.group)

# ============================================================================
# get_field_names() models

class Address(models.Model):
    address_info = models.CharField(max_length=30)


class Course(models.Model):
    course_name = models.CharField(max_length=30)


class Person(models.Model):
    name = models.CharField(max_length=30)
    phone = models.CharField(max_length=30)
    address = models.ForeignKey(Address, on_delete=models.CASCADE)
    courses = models.ManyToManyField(Course)
    best_friend = models.OneToOneField('Person', on_delete=models.CASCADE)


class Building(models.Model):
    superintendent = models.ForeignKey(Person, on_delete=models.CASCADE)





# awl.rankedmodel.views.py

from django.contrib.contenttypes.models import ContentType
from django.contrib.admin.views.decorators import staff_member_required
from django.shortcuts import get_object_or_404
from django.http import HttpResponseRedirect

# ============================================================================

@staff_member_required
def move(request, content_type_id, obj_id, rank):
    """View to be used in the django admin for changing a :class:`RankedModel`
    object's rank.  See :func:`admin_link_move_up` and
    :func:`admin_link_move_down` for helper functions to incoroprate in your
    admin models.

    Upon completion this view sends the caller back to the referring page.

    :param content_type_id:
        ``ContentType`` id of object being moved
    :param obj_id:
        ID of object being moved
    :param rank:
        New rank of the object
    """
    content_type = ContentType.objects.get_for_id(content_type_id)
    obj = get_object_or_404(content_type.model_class(), id=obj_id)
    obj.rank = int(rank)
    obj.save()

    return HttpResponseRedirect(request.META['HTTP_REFERER'])



# awl.managment.commands.run_script.py
#
# runs the python script given on the command line within the django env

import imp
from django.core.management.base import BaseCommand

class Command(BaseCommand):
    """Loads and runs the named python script parameter in this django 
    environment."""

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.help = self.__doc__

    def add_arguments(self, parser):
        parser.add_argument('filename', nargs='+', type=str, 
            help='filename of the python script to run')

    def handle(self, *args, **options):
        imp.load_source('config', options['filename'][0])



import django
from django.contrib import admin

from awl.rankedmodel import urls as ranked_urls
from awl.tests import views

# django 1.11 and 2.0 have different URL path mechanisms 
if django.VERSION[0] >= 2:
    from django.urls import include, path

    urlpatterns = [
        path('admin/', admin.site.urls),

        path('rankedmodel/', include(ranked_urls)),

        path('awl_test_views/test_view_for_messages/', 
            views.test_view_for_messages, ),
    ]
else: # pragma: no cover
    from django.conf.urls import include, url

    urlpatterns = [
        url(r'^admin/', admin.site.urls),

        url(r'^rankedmodel/', include(ranked_urls)),

        url(r'^awl_test_views/test_view_for_messages/$', 
            views.test_view_for_messages, ),
    ]



# awl.tests.test_utils.py
import sys
from django.test import TestCase

from six import StringIO

from awl.tests.models import Link
from awl.utils import (URLTree, refetch, refetch_for_update, render_page,
    render_page_to_string, get_field_names, get_obj_attr)
from awl.waelsteng import FakeRequest

# ============================================================================

class UtilsTest(TestCase):
    def test_url_tree(self):
        # print_tree() exercises everything, so run it and capture stdout
        tree = URLTree()

        saved_stdout = sys.stderr
        try:
            out = StringIO()
            sys.stdout = out
            tree.print_tree()

        finally:
            sys.stdout = saved_stdout

    def test_refetch(self):
        link = Link.objects.create(url='url', text='text')
        link.text = 'foo'

        link = refetch(link)
        self.assertEqual('url', link.url)
        self.assertEqual('text', link.text)

        link.text = 'foo'
        link = refetch_for_update(link)
        self.assertEqual('url', link.url)
        self.assertEqual('text', link.text)

    def test_renders(self):
        request = FakeRequest()
        expected = 'Hello World\n'

        result = render_page_to_string(request, 'sample.html', {'name':'World'})
        self.assertEqual(expected, result)

        response = render_page(request, 'sample.html', {'name':'World'})
        self.assertEqual(expected, response.content.decode('ascii'))

    def test_get_field_names(self):
        from awl.tests.models import Person

        # test defaults, ignore order
        expected = ['name', 'phone']
        result = get_field_names(Person)
        self.assertEqual(set(result), set(expected))

        # test ignore_auto, ignore_relations and exclude
        expected.extend(['id', 'building', 'address', 'courses', 'best_friend',
            'person'])
        expected.remove('phone')
        result = get_field_names(Person, ignore_auto=False,
            ignore_relations=False, exclude=['phone'])
        self.assertEqual(set(result), set(expected))

    def test_get_obj_attr(self):

        # --- data for testing
        class Character(object):
            pass

        class Cartoon(object):
            pass

        barney = Character()
        barney.name = 'Barney'
        betty = Character()
        betty.name = 'Betty'
        betty.husband = barney
        wilma = Character()
        wilma.name = 'Wilma'
        wilma.friend = betty

        cartoon = Cartoon()
        cartoon.name = 'Flinstones'
        cartoon.character = wilma

        # --- tests
        self.assertEqual('Flinstones', get_obj_attr(cartoon, 'name'))
        self.assertEqual(wilma, get_obj_attr(cartoon, 'character'))
        self.assertEqual(betty, get_obj_attr(cartoon, 'character__friend'))
        self.assertEqual(barney, get_obj_attr(cartoon,
            'character__friend__husband'))

        with self.assertRaises(AttributeError):
            get_obj_attr(cartoon, 'foo')

        with self.assertRaises(AttributeError):
            get_obj_attr(cartoon, 'character__foo')



# awl.utils.py

import django
from django.shortcuts import render
from django.template.loader import render_to_string

# =============================================================================
# Template Methods
# =============================================================================

def render_page(request, page_name, data={}):
    """
    .. deprecated:: 0.12
        Use ``django.shortcuts.render`` instead
    
    This function was a wrapper for ``render_to_response`` that handled
    request context.  The ``django.shortcuts.render`` method does the same
    thing, so this just wraps that now. 
    """
    return render(request, page_name, data)


def render_page_to_string(request, page_name, data={}):
    """A shortcut for using ``render_to_string`` with a
    :class:`RequestContext` automatically.
    """
    return render_to_string(page_name, data, request=request)

# ============================================================================
# Object Model Tools
# ============================================================================

def refetch(obj):
    """Queries the database for the same object that is passed in, refetching
    its contents in case they are stale.

    :param obj:
        Object to refetch

    :returns:
        Refreshed version of the object
    """
    return obj.__class__.objects.get(id=obj.id)


def refetch_for_update(obj):
    """Queries the database for the same object that is passed in, refetching
    its contents and runs ``select_for_update()`` to lock the corresponding
    row until the next commit.

    :param obj:
        Object to refetch
    :returns:
        Refreshed version of the object
    """
    return obj.__class__.objects.select_for_update().get(id=obj.id)


def get_field_names(obj, ignore_auto=True, ignore_relations=True, 
        exclude=[]):
    """Returns the field names of a Django model object.

    :param obj: the Django model class or object instance to get the fields
        from
    :param ignore_auto: ignore any fields of type AutoField. Defaults to True
    :param ignore_relations: ignore any fields that involve relations such as
        the ForeignKey or ManyToManyField
    :param exclude: exclude anything in this list from the results

    :returns: generator of found field names
    """

    from django.db.models import (AutoField, ForeignKey, ManyToManyField, 
        ManyToOneRel, OneToOneField, OneToOneRel)

    for field in obj._meta.get_fields():
        if ignore_auto and isinstance(field, AutoField):
            continue

        if ignore_relations and (isinstance(field, ForeignKey) or
                isinstance(field, ManyToManyField) or
                isinstance(field, ManyToOneRel) or
                isinstance(field, OneToOneRel) or
                isinstance(field, OneToOneField)):
            # optimization is killing coverage measure, have to put no-op that
            # does something
            a = 1; a
            continue

        if field.name in exclude:
            continue

        yield field.name


def get_obj_attr(obj, attr):
    """Works like getattr() but supports django's double underscore object
    dereference notation.

    Example usage:

    .. code-block:: python

        >>> get_obj_attr(book, 'writer__age')
        42
        >>> get_obj_attr(book, 'publisher__address')
        <Address object at 105a79ac8>

    :param obj: 
        Object to start the derference from

    :param attr:
        String name of attribute to return

    :returns:
        Derferenced object 

    :raises:
        AttributeError in the attribute in question does not exist
    """
    # handle '__' referencing like in QuerySets
    fields = attr.split('__')
    field_obj = getattr(obj, fields[0])

    for field in fields[1:]:
        # keep going down the reference tree
        field_obj = getattr(field_obj, field)

    return field_obj

# ============================================================================
# Misc
# ============================================================================

class URLTree(object):
    """A tree representation of the django url paths.  Each path is stored in
    a dictionary with its regex pattern, full path, name (if available) and
    children.  Root items in the tree can be accessed through the ``children``
    list attribute.
    """
    def __init__(self):
        self.children = []

        # parse the url pattern tree 
        from django.conf import settings
        urlconf = __import__(settings.ROOT_URLCONF, {}, {}, [''])
        self._parse_node(urlconf.urlpatterns, '', self.children)

    def _parse_node(self, parent, parent_path, children):
        for entry in parent:
            # django 1.11 and 2.0 have different URLResolver objects
            if django.VERSION[0] >= 2:
                pattern = entry.pattern.regex.pattern
            else:  # pragma: no cover
                pattern = entry.regex.pattern

            path = parent_path + pattern
            d = {
                'pattern':pattern,
                'path':path,
                'name':entry.name if hasattr(entry, 'name') else '',
                'children':[],
            }
            children.append(d)

            if hasattr(entry, 'url_patterns'):
                self._parse_node(entry.url_patterns, path, d['children'])

    def _depth_traversal(self, node, result):
        text = node['path']
        if node['name']:
            text += ', name=%s' % node['name']

        result.append(text)
        for child in node['children']:
            self._depth_traversal(child, result)

    def as_list(self):
        """Returns a list of strings describing the full paths and patterns
        along with the name of the urls.  Example:

        .. code-block::python
            >>> u = URLTree()
            >>> u.as_list()
            [
                'admin/',
                'admin/$, name=index',
                'admin/login/$, name=login',
            ]
        """
        result = []
        for child in self.children:
            self._depth_traversal(child, result)

        return result

    def print_tree(self):
        """Convenience method for printing the results of
        :class:`URLTree.as_list` to STDOUT
        """
        for line in self.as_list():
            print(line)



# awl.tests.test_contextproc.py

from django.test import TestCase

from awl.context_processors import extra_context
from awl.waelsteng import FakeRequest

# ============================================================================

class ContextProcessorTest(TestCase):
    def test_extra_context(self):
        request = FakeRequest()
        context = extra_context(request)

        self.assertEqual(request, context['request'])
        self.assertEqual('test_host', context['HOST'])
        self.assertFalse(context['IN_ADMIN'])
